function []=lab4_demo()
%   Demonstration of edge enhancement of blurred and 
%   blurred & noisy image using Inverse Filtering,
%   Pseudo Inverse Filtering and Wiener Filtering.
%   The test images are checkerboard and fragment of Lena image.

%--------------------------------------------------------------------------
% Test image - checkerboard
%--------------------------------------------------------------------------
disp('-----------------------------------------------------------');
disp('Test image - checkerboard');
% Load checkerboard
f=checkerboard(8);

% Initialization
[M,N]=size(f);
w1=4;
w2=4;
h=fspecial('gaussian',[w1 w2],0.5);
h_padded=zeros(M,N);
h_padded(M/2-w1/2+1:M/2+w1/2,N/2-w2/2+1:N/2+w2/2)=h; % Zero padding of the 
                        % frequency response of the filter to the size MxN.
H=fftshift(fft2(h_padded));
blurred=imfilter(f,h,'symmetric','same','conv');
Blurred=fftshift(fft2(blurred));
variance=0.01;
blurred_noisy=imnoise(blurred,'gaussian',0,variance);
Blurred_noisy=fftshift(fft2(blurred_noisy));

figure;
subplot(2,2,1);
imshow(f,[]);
title('The initial image f(i,j)');
subplot(2,2,2);
imshow(h,[]);
axis on;
title('The blurring filter h(i,j)');
subplot(2,2,3);
imshow(blurred,[]);
title('The blurred image w(i,j)');
subplot(2,2,4);
imshow(blurred_noisy,[]);
title('The blurred and noisy image g(i,j)');
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Inverse filtering');
H_I=H.^(-1);
if any(isinf(H_I(:)))
    warning('The Inverse Filter has Inf entries!!!');
else
    disp('The Inverse Filter has no Inf entries.');
end

%   Edge enhancement of blurred image
F_enhanced_I1=Blurred.*H_I;
f_enhanced_I1=abs(ifftshift(ifft2(F_enhanced_I1)));
if any(isnan(f_enhanced_I1(:)))
    warning('The image with enhanced edges has NaN entries!!!');
else
    disp('The image with enhanced edges has no NaN entries.');
end
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Pseudo Inverse Filtering');
epsilon=1.2e-1;
H_PI=zeros(M,N);
index=find(abs(H)>epsilon);
H_PI(index)=1./H(index);
if any(isinf(H_PI(:)))
    warning('The Pseudo Inverse Filter has Inf entries!!!');
else
    disp('The Pseudo Inverse Filter has no Inf entries.');
end

%   Edge enhancement of blurred image
F_enhanced_PI1=Blurred.*H_PI;
f_enhanced_PI1=abs(ifftshift(ifft2(F_enhanced_PI1)));
if any(isnan(f_enhanced_PI1(:)))
    disp('The image with enhanced edges has NaN entries!!!');
else
    disp('The image with enhanced edges has no NaN entries.');
end

%   Edge enhancement of blurred & noisy image
F_enhanced_PI2=Blurred_noisy.*H_PI;
f_enhanced_PI2=abs(ifftshift(ifft2(F_enhanced_PI2)));

figure;
subplot(2,2,1);
imshow(abs(H_PI),[]);
title('Pseudo Inverse Filter H^{PI}(u,v)');
subplot(2,2,3);
imshow(f_enhanced_PI1,[]);
title('Pseudo Inv. edge enh. f_w(i,j) of blurr. img');
subplot(2,2,4);
imshow(f_enhanced_PI2,[]);
title('Pseudo Inv. edge enh. f_g(i,j) of blurr.&noisy img');
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Wiener Filtering');
alpha=255*0.001;
[v,u]=meshgrid(-N/2:N/2-1,-M/2:M/2-1);

%   Edge enhancement of blurred image
variance1=0;
denom1=abs(H).^2+alpha*variance1*(u.^2+v.^2);
[j1,j2]=find(denom1==0);
denom1(j1,j2)=eps; % Prevent division by zero for pixels where abs(H)=0
H_WI1=conj(H)./denom1;
if any(isinf(H_WI1(:)))
    warning('The Wiener Filter has Inf entries!!!');
else
    disp('The Wiener Filter has no Inf entries.');
end
F_enhanced_WI1=Blurred.*H_WI1;
f_enhanced_WI1=abs(ifftshift(ifft2(F_enhanced_WI1)));
if any(isnan(f_enhanced_WI1(:)))
    disp('The enhanced image has NaN entries!!!');
else
    disp('The enhanced image has no NaN entries.');
end

%   Edge enhancement of blurred & noisy image
variance2=variance; % =0.01 in our case, as defined 
                    % in the Initialization section of this program
H_WI2=conj(H)./(abs(H).^2+alpha*variance2*(u.^2+v.^2));
F_enhanced_WI2=Blurred_noisy.*H_WI2;
f_enhanced_WI2=abs(ifftshift(ifft2(F_enhanced_WI2)));

figure;
subplot(2,2,1);
imshow(abs(H_WI1),[]);
title('Wiener Filter H^{WI}(u,v) for deblurring');
subplot(2,2,2);
imshow(abs(H_WI2),[]);
title('Wiener Filter H^{WI}(u,v) for deblurring & denoising');
subplot(2,2,3);
imshow(f_enhanced_WI1,[]);
title('Wiener edge enh. f_w(i,j) of blurred img');
subplot(2,2,4);
imshow(f_enhanced_WI2,[]);
title('Wiener edge enh. f_g(i,j) of blurred & noisy img');
disp('Press any key to continue.');
pause;

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
% Test image -fragment of Lena
disp('-----------------------------------------------------------');
disp('-----------------------------------------------------------');
disp('Test image - fragment of Lena');
disp('-----------------------------------------------------------');
disp('-----------------------------------------------------------');
% Load fragment of Lena
f=imread('lena.gif');
f=im2double(f(201:300,201:300));

% Initialization
[M,N]=size(f);
w1=4;
w2=4;
h=fspecial('gaussian',[w1 w2],0.5);
h_padded=zeros(M,N);
h_padded(M/2-w1/2+1:M/2+w1/2,N/2-w2/2+1:N/2+w2/2)=h; % Zero padding of the 
                        % frequency response of the filter to the size MxN.
H=fftshift(fft2(h_padded));
blurred=imfilter(f,h,'symmetric','same','conv');
Blurred=fftshift(fft2(blurred));
variance=0.01;
blurred_noisy=imnoise(blurred,'gaussian',0,variance);
Blurred_noisy=fftshift(fft2(blurred_noisy));

figure;
subplot(2,2,1);
imshow(f,[]);
title('The initial image f(i,j)');
subplot(2,2,2);
imshow(h,[]);
axis on;
title('The blurring filter h(i,j)');
subplot(2,2,3);
imshow(blurred,[]);
title('The blurred image w(i,j)');
subplot(2,2,4);
imshow(blurred_noisy,[]);
title('The blurred and noisy image g(i,j)');
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Inverse filtering');
H_I=H.^(-1);
if any(isinf(H_I(:)))
    warning('The Inverse Filter has Inf entries!!!');
else
    disp('The Inverse Filter has no Inf entries.');
end

%   Edge enhancement of blurred image
F_enhanced_I1=Blurred.*H_I;
f_enhanced_I1=abs(ifftshift(ifft2(F_enhanced_I1)));
if any(isnan(f_enhanced_I1(:)))
    warning('The image with enhanced edge has NaN entries!!!');
else
    disp('The image with enhanced edge has no NaN entries.');
end
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Pseudo Inverse Filtering');
epsilon=1.2e-1;
H_PI=zeros(M,N);
index=find(abs(H)>epsilon);
H_PI(index)=1./H(index);
if any(isinf(H_PI(:)))
    warning('The Pseudo Inverse Filter has Inf entries!!!');
else
    disp('The Pseudo Inverse Filter has no Inf entries.');
end

%   Edge enhancement of blurred image
F_enhanced_PI1=Blurred.*H_PI;
f_enhanced_PI1=abs(ifftshift(ifft2(F_enhanced_PI1)));
if any(isnan(f_enhanced_PI1(:)))
    disp('The image with enhanced edges has NaN entries!!!');
else
    disp('The image with enhanced edges has no NaN entries.');
end

%   Edge enhancement of blurred & noisy image
F_enhanced_PI2=Blurred_noisy.*H_PI;
f_enhanced_PI2=abs(ifftshift(ifft2(F_enhanced_PI2)));

figure;
subplot(2,2,1);
imshow(abs(H_PI),[]);
title('Pseudo Inverse Filter H^{PI}(u,v)');
subplot(2,2,3);
imshow(f_enhanced_PI1,[]);
title('Pseudo Inv. edge enh. f_w(i,j) of blurr. img');
subplot(2,2,4);
imshow(f_enhanced_PI2,[]);
title('Pseudo Inv. edge enh. f_g(i,j) of blurr.&noisy img');
disp('Press any key to continue.');
pause;

disp('-----------------------------------------------------------');
disp('Wiener Filtering');
alpha=255*0.001;
[v,u]=meshgrid(-N/2:N/2-1,-M/2:M/2-1);

%   Edge enhancement of blurred image
variance1=0;
denom1=abs(H).^2+alpha*variance1*(u.^2+v.^2);
[j1,j2]=find(denom1==0);
denom1(j1,j2)=eps; % The correction in order to prevent division by zero 
                   % for pixels where abs(H)=0.
H_WI1=conj(H)./denom1;
if any(isinf(H_WI1(:)))
    warning('The Wiener Filter has Inf entries (after correction)!!!');
else
    disp('The Wiener Filter has no Inf entries (after correction).');
end
F_enhanced_WI1=Blurred.*H_WI1;
f_enhanced_WI1=abs(ifftshift(ifft2(F_enhanced_WI1)));
if any(isnan(f_enhanced_WI1(:)))
    disp('The enhanced image has NaN entries (after correction)!!!');
else
    disp('The enhanced image has no NaN entries (after correction).');
end

%   Edge enhancement of blurred & noisy image
variance2=variance; % =0.01 in our case, as defined 
                    % in the Initialization section of this program
H_WI2=conj(H)./(abs(H).^2+alpha*variance2*(u.^2+v.^2));
F_enhanced_WI2=Blurred_noisy.*H_WI2;
f_enhanced_WI2=abs(ifftshift(ifft2(F_enhanced_WI2)));

figure;
subplot(2,2,1);
imshow(abs(H_WI1),[]);
title('Wiener Filter H^{WI}(u,v) for deblurring');
subplot(2,2,2);
imshow(abs(H_WI2),[]);
title('Wiener Filter H^{WI}(u,v) for deblurring & denoising');
subplot(2,2,3);
imshow(f_enhanced_WI1,[]);
title('Wiener edge enh. f_w(i,j) of blurred img');
subplot(2,2,4);
imshow(f_enhanced_WI2,[]);
title('Wiener edge enh. f_g(i,j) of blurred & noisy img');